#include <stdio.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/bitmaps/gray>
#include <math.h>
#include "xlfdutil.h"

static int
compare_int(x, y)
int *x, *y;
{
    return *x - *y;
}

char *
do_charset(string)
char *string;
{
    static char result[1024];
    int i, count = strlen(string), last;
    int *charlist = (int *)malloc(sizeof(int) * count);

    if (charlist == NULL)
    {
	fprintf(stderr, "malloc() failed\n");
	exit(1);
    }

    for (i = 0; i < count; i++) charlist[i] = string[i];
    qsort((char *)charlist,
	  count,
	  sizeof(int),
	  compare_int);

    last = charlist[0];
    sprintf(result, "%d", last);
    for (i = 1; i < count; i++)
	if (charlist[i] != last)
	{
	    last = charlist[i];
	    sprintf(result + strlen(result), ",%d", last);
	}
    free(charlist);
    return result;
}

main(argc, argv)
int argc;
char *argv[];
{
    Display *display;
    Window window;
    GC gc1, gc2;
    XGCValues values;
    int i, d1count, d2count, d4count, d4size, d4size2, screen;
    XFontStruct *d1fonts[30];
    XFontStruct *d2fonts[20];
    XFontStruct *d3fonts[4];
    XFontStruct *d4fonts[20];
    XFontStruct *tempfont;
    char *d1string = "Artificial Obliquing";
    char *d2string = "Anamorphic Scaling";
    char *d3string1 = "Mirroring";
    char *d3string2 = "gnirorriM";
    char *d4string = "Glyph Rotation";
    char buffer[10];
    Pixmap gray;
    double y, div, ln2 = log((double)2.0);
    int junk;
    XCharStruct d1stats, d3stats;
    double radius;
    XLFDRec Record;
    char *fn;

    if (argc < 2 || ParseXLFD(argv[1], &Record, FONT_XLFD_REPLACE_NONE) == NULL)
    {
	fprintf(stderr, "Usage: %s <valid_XLFD_name>\n", argv[0]);
	exit(1);
    }

    display = XOpenDisplay(NULL);
    screen = DefaultScreen(display);
    window = XCreateSimpleWindow(display,
				 DefaultRootWindow(display),
				 0, 0, 600, 600, 0,
				 WhitePixel(display, screen),
				 WhitePixel(display, screen));
    XSelectInput(display, window, ExposureMask);
    XMapWindow(display, window);

    values.background = WhitePixel(display, screen);
    values.foreground = BlackPixel(display, screen);
    values.line_width = 0;
    gc1 = XCreateGC(display, window, GCForeground | GCBackground, &values);

    gray = XCreateBitmapFromData(display, window, gray_bits, gray_width,
				 gray_height);
    values.fill_style = FillStippled;
    values.stipple = gray;
    gc2 = XCreateGC(display, window,
		    GCForeground | GCBackground | GCFillStyle | GCStipple,
		    &values);

    Record.pixelSize = 60;
    Record.pointSize = 0;
    Record.resolutionX = 100;
    Record.resolutionY = 100;
    Record.averageWidth = 0;
    Record.pixelSetsize = 60;
    Record.pointSetsize = 0;
    Record.slantAngle = 0;
    Record.mirrorX = False;
    Record.mirrorY = False;
    Record.rotationAngle = 0;
    Record.ranges = do_charset(d1string);
    d1fonts[0] = XLoadQueryFont(display,
		    fn = ParseXLFD(NULL, &Record, FONT_XLFD_REPLACE_VALUE));
    if (d1fonts[0] == NULL)
    {
	fprintf(stderr, "Cannot open font \"%s\"\n", fn);
	exit(1);
    }
    free(fn);

    d1count = strlen(d1string);
    XTextExtents(d1fonts[0], d1string, strlen(d1string),
		 &junk, &junk, &junk, &d1stats);
    y = (double)d1stats.width / 2.0;

    Record.ranges = buffer;
    Record.mirrorY = True;
    for (i = 0; i < d1count; i++)
    {
	double x = (double)XTextWidth(d1fonts[0], d1string, i) +
		   (double)XTextWidth(d1fonts[0], d1string + i, 1) / 2.0;
	Record.slantAngle = atan2(y - x, y) * 11520.0 / M_PI;
	sprintf(buffer, "%d", d1string[i]);
	d1fonts[i + 1] = XLoadQueryFont(display,
		    fn = ParseXLFD(NULL, &Record, FONT_XLFD_REPLACE_VALUE));
	if (d1fonts[i + 1] == NULL)
	{
	    fprintf(stderr, "Cannot open font \"%s\"\n", fn);
	    exit(1);
	}
	free(fn);
    }
    Record.slantAngle = 0;
    Record.mirrorY = False;

    d2count = strlen(d2string);
    div = (double)(d2count - 1);
    for (i = 0; i < d2count; i++)
    {
	double mult = exp(ln2 * (1.0 - 2.0 * (double)i / div));
	Record.pixelSize = 50.0 * mult;
	Record.pixelSetsize = 50.0 / mult;
	sprintf(buffer, "%d", d2string[i]);
	d2fonts[i] = XLoadQueryFont(display,
		    fn = ParseXLFD(NULL, &Record, FONT_XLFD_REPLACE_VALUE));
	if (d2fonts[i] == NULL)
	{
	    fprintf(stderr, "Cannot open font \"%s\"\n", fn);
	    exit(1);
	}
	free(fn);
    }
    Record.pixelSize = Record.pixelSetsize = 60;

    Record.ranges = do_charset(d3string1);
    d3fonts[0] = XLoadQueryFont(display,
		    fn = ParseXLFD(NULL, &Record, FONT_XLFD_REPLACE_VALUE));
    if (d3fonts[0] == NULL)
    {
	fprintf(stderr, "Cannot open font \"%s\"\n", fn);
	exit(1);
    }
    free(fn);

    Record.mirrorX = True;
    d3fonts[1] = XLoadQueryFont(display,
		    fn = ParseXLFD(NULL, &Record, FONT_XLFD_REPLACE_VALUE));
    if (d3fonts[1] == NULL)
    {
	fprintf(stderr, "Cannot open font \"%s\"\n", fn);
	exit(1);
    }
    free(fn);

    Record.mirrorX = False;
    Record.mirrorY = True;
    d3fonts[2] = XLoadQueryFont(display,
		    fn = ParseXLFD(NULL, &Record, FONT_XLFD_REPLACE_VALUE));
    if (d3fonts[2] == NULL)
    {
	fprintf(stderr, "Cannot open font \"%s\"\n", fn);
	exit(1);
    }
    free(fn);

    Record.mirrorX = True;
    d3fonts[3] = XLoadQueryFont(display,
		    fn = ParseXLFD(NULL, &Record, FONT_XLFD_REPLACE_VALUE));
    if (d3fonts[3] == NULL)
    {
	fprintf(stderr, "Cannot open font \"%s\"\n", fn);
	exit(1);
    }
    free(fn);

    XTextExtents(d3fonts[0], d3string1, strlen(d3string1),
		 &junk, &junk, &junk, &d3stats);

    Record.mirrorX = Record.mirrorY = False;
    Record.ranges = do_charset(d4string);
    tempfont = XLoadQueryFont(display,
		    fn = ParseXLFD(NULL, &Record, FONT_XLFD_REPLACE_VALUE));
    if (tempfont == NULL)
    {
	fprintf(stderr, "Cannot open font \"%s\"\n", fn);
	exit(1);
    }
    free(fn);

    d4count = strlen(d4string);
    d4size = XTextWidth(tempfont, d4string, d4count);
    d4size2 = 0;
    div = (double)d4size;
    radius = div / M_PI;
    Record.ranges = buffer;
    for (i = 0; i < d4count; i++)
    {
	int charsize = XTextWidth(tempfont, d4string + i, 1);
	double angle = 5760.0 -
		       (double)(d4size2 + charsize / 2) / div * 11520.0;
	Record.rotationAngle = angle;
	sprintf(buffer, "%d", d4string[i]);
	d4fonts[i] = XLoadQueryFont(display,
		       fn = ParseXLFD(NULL, &Record, FONT_XLFD_REPLACE_VALUE));
	if (d4fonts[i] == NULL)
	{
	    fprintf(stderr, "Cannot open font \"%s\"\n", fn);
	    exit(1);
	}
	free(fn);
	d4size2 += charsize;
    }
    XFreeFont(display, tempfont);

    for (;;)
    {
	XEvent event;
	XNextEvent(display, &event);
	if (event.type == Expose && event.xexpose.count == 0)
	{
	    int location;
	    int xcenter, ycenter;
	    char *p1, *p2;

	    values.font = d1fonts[0]->fid;
	    XChangeGC(display, gc1, GCFont, &values);
	    XDrawString(display, window, gc1, 23, 80, d1string,
			strlen(d1string));

	    for (i = 0; i < d1count; i++)
	    {
		values.font = d1fonts[i + 1]->fid;
		XChangeGC(display, gc2, GCFont, &values);
		XDrawString(display, window, gc2,
			    23 + XTextWidth(d1fonts[0], d1string, i),
			    100,
			    d1string + i,
			    1);
	    }

	    location = 50;
	    for (i = 0; i < d2count; i++)
	    {
		values.font = d2fonts[i]->fid;
		XChangeGC(display, gc1, GCFont, &values);
		XDrawString(display, window, gc1,
			    location,
			    210,
			    d2string + i,
			    1);
		location += XTextWidth(d2fonts[i], d2string + i, 1);
	    }

	    values.font = d3fonts[0]->fid;
	    XChangeGC(display, gc1, GCFont, &values);
	    XDrawString(display, window, gc1, 25, 280, d3string1,
			strlen(d3string1));

	    values.font = d3fonts[1]->fid;
	    XChangeGC(display, gc1, GCFont, &values);
	    XDrawString(display, window, gc1, 45 + d3stats.width,
			280, d3string2, strlen(d3string2));

	    values.font = d3fonts[2]->fid;
	    XChangeGC(display, gc1, GCFont, &values);
	    XDrawString(display, window, gc1, 25,
			290 + 2 * d3stats.descent,
			d3string1,
			strlen(d3string1));

	    values.font = d3fonts[3]->fid;
	    XChangeGC(display, gc1, GCFont, &values);
	    XDrawString(display, window, gc1, 45 + d3stats.width,
			290 + 2 * d3stats.descent,
			d3string2,
			strlen(d3string2));

	    XDrawLine(display, window, gc1,
		      20,
		      285 + d3stats.descent,
		      50 + 2 * d3stats.width,
		      285 + d3stats.descent);

	    XDrawLine(display, window, gc1,
		      35 + d3stats.width,
		      285 - d3stats.ascent,
		      35 + d3stats.width,
		      285 + d3stats.ascent + 2 * d3stats.descent);

	    xcenter = 300;
	    ycenter = 550;
	    d4size2 = 0;
	    for (i = 0; i < d4count; i++)
	    {
		double angle = (double)(d4size - d4size2) /
			       (double)d4size * M_PI;
		values.font = d4fonts[i]->fid;
		XChangeGC(display, gc1, GCFont, &values);
		XDrawString(display, window, gc1,
			    xcenter + (int)(radius * cos(angle)),
			    ycenter - (int)(radius * sin(angle)),
			    d4string + i, 1);
		d4size2 += XTextWidth(d4fonts[i], d4string + i, 1);
	    }

	}
    }
}
