#include <stdio.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/X.h>
#include <X11/cursorfont.h>
#include <X11/Xutil.h>
#include "xlfdutil.h"

/* Generate a display of characters rotated at a range of angles.

    Usage:

      $0 <XLFD-name> <start-angle> <end-angle> <angle-increment>

    Example:

      $0 "-adobe-utopia-medium-r-normal--0-180+120-0-0-p-0-iso8859-1" 0 330 30
*/

char *progname;

usage()
{
  int i,j,spaces,xloc;
  static char *options[] = {
    "[-display <displayname>]",
    "[-geometry <geometry>]",
    "<typeface>",
    "<start-angle>",
    "<end-angle>",
    "<angle-increment>",
  };
  (void)fprintf(stderr,"Usage: %s", progname);
  spaces=strlen(progname)+7;
  xloc=spaces;

  for (j=0; j<(sizeof(options)/sizeof(char *)); j++) {
    if (xloc+strlen(options[j]) > 78) {
      putc('\n',stderr);
      for (i = 0; i<spaces; i++) (void)putc(' ',stderr);
      xloc=spaces;
    }
    xloc += strlen(options[j])+1;
    fprintf(stderr," %s",options[j]);
  }
  putc('\n',stderr);

  exit(1);
}

main(argc, argv)
int argc;
char *argv[];
{
  double angle, anglelo = -1.0, anglehi = -1.0, angleinc = -1.0;
  char fontname[1024], outbuffer[1024], *ptr;
  char *xlfdstring = NULL, *geomstring = NULL;
  char *displayname = NULL;
  int width, height;
  unsigned long blackpixel, whitepixel;
  Display *display;
  int i, j, screen;
  Window window, rootwindow;
  Pixmap pixmap;
  Cursor cursor;
  XSetWindowAttributes attributes;
  XSizeHints size_hints;
  GC windowgc, pixmapgc;
  XGCValues values;
  XEvent event;
  XFontStruct *fontinfo;
  int nargc = argc;
  char **nargv = argv;
  XLFDRec Record;

  progname = argv[0];

  while (--nargc)
  {
    if (!strcmp(*++nargv, "-display"))
    {
      if (--nargc == 0) usage();
      displayname = *++nargv;
    }
    else if (!strcmp(*nargv, "-geometry"))
    {
      if (--nargc == 0) usage();
      geomstring = *++nargv;
    }
    else if (xlfdstring == NULL)
      xlfdstring = *nargv;
    else if (anglelo == -1.0)
    {
      anglelo = atof(*nargv);
    }
    else if (anglehi == -1.0)
    {
      anglehi = atof(*nargv);
    }
    else if (angleinc == -1.0)
    {
      angleinc = atof(*nargv);
    }
    else usage();
  }

  if (anglehi < anglelo || angleinc <= 0.0) usage();

  if (ParseXLFD(xlfdstring, &Record, FONT_XLFD_REPLACE_NONE) == NULL)
  {
    fprintf(stderr, "%s: Font name \"%s\" invalid\n", progname, xlfdstring);
    exit(1);
  }

  display = XOpenDisplay(displayname);
  if (display == NULL)
  {
    fprintf(stderr, "%s: Cannot open display \"%s\"\n", progname, displayname);
    exit(1);
  }
  screen = DefaultScreen(display);
  rootwindow = RootWindow(display, screen);
  blackpixel = BlackPixel(display, screen);
  whitepixel = WhitePixel(display, screen);

  if (display == NULL)
  {
    fprintf(stderr, "%s: XOpenDisplay failed\n", progname);
    exit(1);
  }

  cursor = XCreateFontCursor(display, XC_watch);

  height = 1000;
  width = 1000;
  size_hints.x = 0;
  size_hints.y = 0;
  size_hints.width = width;
  size_hints.height = height;
  size_hints.flags = PPosition | PSize;

  if (geomstring)
  {
    int result;
    result = XParseGeometry(geomstring,&size_hints.x,&size_hints.y,
			    &size_hints.width,&size_hints.height);
    if (result & XValue || result & YValue)
    {
      size_hints.flags |= USPosition;
      size_hints.flags &= ~PPosition;
    }
    if (result & WidthValue || result & HeightValue)
    {
      size_hints.flags |= USSize;
      size_hints.flags &= ~PSize;
    }

    if (result & WidthValue)
      width = size_hints.width;

    if (result & HeightValue)
      height = size_hints.height;

    if (result & XNegative)
      size_hints.x += DisplayWidth(display,screen) - size_hints.width;
    if (result & YNegative)
      size_hints.y += DisplayHeight(display,screen) - size_hints.height;
  }

  attributes.background_pixel = whitepixel;
  attributes.border_pixel = blackpixel;
  attributes.event_mask = ExposureMask | KeyPressMask;
  attributes.cursor = cursor;

  window = XCreateWindow(display, rootwindow, 0, 0, width, height,
			 2, CopyFromParent, InputOutput, CopyFromParent,
			 CWBackPixel | CWBorderPixel | CWEventMask | CWCursor,
			 &attributes);

  XSetStandardProperties(display, window, xlfdstring, xlfdstring, None,
			 argv, argc, &size_hints);

  values.background = whitepixel;
  values.foreground = blackpixel;
  windowgc = XCreateGC(display, window, GCForeground | GCBackground, &values);

  /* create depth-1 pixmap */
  pixmap = XCreatePixmap(display, rootwindow, width, height, 1);

  values.background = 0;
  values.foreground = 0;
  pixmapgc = XCreateGC(display, pixmap, GCForeground | GCBackground, &values);
  XFillRectangle(display, pixmap, pixmapgc, 0, 0, width, height);
  values.foreground = 1;
  XChangeGC(display, pixmapgc, GCForeground, &values);

  XMapWindow(display, window);
  XSync(display, 0);

  ptr = outbuffer;
  for (i = 33; i < 128; i++) *ptr++ = i;
  for (i = 161; i < 256; i++) *ptr++ = i;
  *ptr = '\0';

  for (angle = anglelo; angle <= anglehi; angle += angleinc)
  {
    char *fontname;

    Record.rotationAngle = (long)(angle * 64.0);
    if (fontname = ParseXLFD(NULL, &Record, FONT_XLFD_REPLACE_VALUE))
    {
      fontinfo = XLoadQueryFont(display, fontname);
      if (fontinfo == NULL)
	fprintf(stderr, "%s: Cannot open font \"%s\"\n", progname, fontname);
      else
      {
	values.font = fontinfo->fid;
	XChangeGC(display, windowgc, GCFont, &values);
	XChangeGC(display, pixmapgc, GCFont, &values);
	XDrawRotatedString(display, window, windowgc, width / 2, height / 2,
			   outbuffer, ptr - outbuffer, fontinfo,
			   Record.rotationAngle);
	XDrawRotatedString(display, pixmap, pixmapgc, width / 2, height / 2,
			   outbuffer, ptr - outbuffer, fontinfo,
			   Record.rotationAngle);
	XFreeFont(display, fontinfo);
      }
      free(fontname);
    }

  }

  attributes.cursor = None;
  XChangeWindowAttributes(display, window, CWCursor, &attributes);

  for (;;)
  {
    XNextEvent(display, &event);
    if (event.type == Expose)
    {
      XCopyPlane(display, pixmap, window, windowgc,
                 event.xexpose.x, event.xexpose.y,
		 event.xexpose.width, event.xexpose.height,
                 event.xexpose.x, event.xexpose.y, 0x1);
    }
    else if (event.type == KeyPress)
    {
      char keybuffer[10];
      int len;
      len = XLookupString(&event.xkey, keybuffer, sizeof(keybuffer), NULL,
			  NULL);
      if (len == 1 && (keybuffer[0] == 'q' ||
		       keybuffer[0] == 'Q' ||
		       keybuffer[0] == '\003'))
      {
	XCloseDisplay(display);
	exit(0);
      }
    }
  }

  /*NOTREACHED*/
}
